(* $Id: base64.ml,v 1.2 1999/06/27 11:22:34 gerd Exp $
 * ----------------------------------------------------------------------
 *
 *)

let b64_pattern plus slash =
  [| 'A'; 'B'; 'C'; 'D'; 'E'; 'F'; 'G'; 'H'; 'I'; 'J'; 'K'; 'L'; 'M';
     'N'; 'O'; 'P'; 'Q'; 'R'; 'S'; 'T'; 'U'; 'V'; 'W'; 'X'; 'Y'; 'Z';
     'a'; 'b'; 'c'; 'd'; 'e'; 'f'; 'g'; 'h'; 'i'; 'j'; 'k'; 'l'; 'm';
     'n'; 'o'; 'p'; 'q'; 'r'; 's'; 't'; 'u'; 'v'; 'w'; 'x'; 'y'; 'z';
     '0'; '1'; '2'; '3'; '4'; '5'; '6'; '7'; '8'; '9'; plus; slash |]
;;


let encode_special plus slash equal s =
  (* encode using "base64".
   * 'plus': The character that should be used instead of '+' in the original
   *         encoding scheme. Pass '+' to get the original encoding scheme.
   * 'slash': The character that should be used instead of '/' in the original
   *          encoding scheme. Pass '/' to get the original encoding scheme.
   * 'equal': The character that should be used instead of '=' in the original
   *          encoding scheme. Pass '=' to get the original encoding scheme.
   * 's': The string to be encoded.
   *)
  let b64 = b64_pattern plus slash in
  let l_s = String.length s in
  let l_t = if l_s = 0 then 0 else ((l_s - 1) / 3 + 1) * 4 in
  let t = String.make l_t equal in
  for k = 0 to l_s / 3 - 1 do
    let bits = (Char.code (s.[3*k])   lsl 16) lor
               (Char.code (s.[3*k+1]) lsl  8) lor
               (Char.code (s.[3*k+2])) in
    t.[4*k]   <- b64.( bits lsr 18);
    t.[4*k+1] <- b64.((bits lsr 12) land 0x3f);
    t.[4*k+2] <- b64.((bits lsr  6) land 0x3f);
    t.[4*k+3] <- b64.( bits         land 0x3f)
  done;
  (* padding if needed: *)
  begin
    match l_s mod 3 with
      0 -> ()
    | 1 ->
        let bits = Char.code (s.[l_s - 1]) in
	t.[ l_t - 4 ] <- b64.( bits lsr 2);
	t.[ l_t - 3 ] <- b64.( (bits land 0x03) lsl 4);
    | 2 ->
	let bits = (Char.code (s.[l_s - 2]) lsl 8) lor
                   (Char.code (s.[l_s - 1])) in
	t.[ l_t - 4 ] <- b64.( bits lsr 10);
	t.[ l_t - 3 ] <- b64.((bits lsr  4) land 0x3f);
	t.[ l_t - 2 ] <- b64.((bits lsl  2) land 0x3f);
    | _ -> failwith "this is impossible"
  end;

  t
;;


let encode = encode_special '+' '/' '=';;  
    (* official base 64 encoding *)
let url_encode = encode_special '-' '/' '.';;
    (* use '-' and '.' instead of '+' and '=' in URLs *)

let decode t =
  (* t: may be formed using 'encode' or 'url_encode' *)
  let l_t = String.length t in
  if l_t mod 4 <> 0 then
    failwith "Base64.decode";
  let pad_chars =
    if l_t > 0 then begin
      if String.sub t (l_t - 2) 2 = "==" or String.sub t (l_t - 2) 2 = ".." then
	2
      else if String.sub t (l_t - 1) 1 = "=" or String.sub t (l_t - 1) 1 = "." then
	1
      else
	0
    end
    else 0 in
  let l_s = (l_t / 4) * 3 - pad_chars in       (* sic! *)
  let s = String.create l_s in
  let b' c =
    match c with
      'A' .. 'Z'  -> Char.code(c) - 65     (* 65 = Char.code 'A' *)
    | 'a' .. 'z'  -> Char.code(c) - 71     (* 71 = Char.code 'a' - 26 *)
    | '0' .. '9'  -> Char.code(c) + 4      (* -4 = Char.code '0' - 52 *)
    | ('+' | '-') -> 62
    | '/'         -> 63
    | ('=' | '.') -> 0
    | _           -> failwith "Base64.decode"
  in
  for k = 0 to l_t / 4 - 1 do
    let c0 = t.[ k*4 ] in
    let c1 = t.[ k*4 + 1 ] in
    let c2 = t.[ k*4 + 2 ] in
    let c3 = t.[ k*4 + 3 ] in
    let n0 = b' c0 in
    let n1 = b' c1 in
    let n2 = b' c2 in
    let n3 = b' c3 in
    let x0 = (n0 lsl 2) lor (n1 lsr 4) in
    let x1 = ((n1 lsl 4) land 0xf0) lor (n2 lsr 2) in
    let x2 = ((n2 lsl 6) land 0xc0) lor n3 in
    if c0 = '=' or c0 = '.' or c1 = '=' or c1 = '.' then
      failwith "Base64.decode"
    else
      if c2 = '=' or c2 = '.' then begin
        if (c3 <> '=' & c3 <> '.') or k <> l_t/4 - 1 then
	  failwith "Base64.decode";
        s.[ k*3 ]   <- Char.chr x0;
      end
      else
        if c3 = '=' or c3 = '.' then begin
	  if k <> l_t/4 - 1 then
	    failwith "Base64.decode";
          s.[ k*3 ]   <- Char.chr x0;
	  s.[ k*3+1 ] <- Char.chr x1;
        end
          else begin
	    s.[ k*3 ]   <- Char.chr x0;
	    s.[ k*3+1 ] <- Char.chr x1;
	    s.[ k*3+2 ] <- Char.chr x2;
          end
  done;

  s
;;

(* ======================================================================
 * History:
 * 
 * $Log: base64.ml,v $
 * Revision 1.2  1999/06/27 11:22:34  gerd
 * 	Slightly improved version.
 *
 * Revision 1.1  1999/03/26 00:45:29  gerd
 * 	Initial revision.
 *
 * 
 *)
